/*
 * Copyright 2009-2010 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

/*!
 * @file imx_adc_ts.c
 *
 * @brief Driver for the Freescale Semiconductor i.MX ADC touchscreen.
 *
 * This touchscreen driver is designed as a standard input driver.  It is a
 * wrapper around the low level ADC driver. Much of the hardware configuration
 * and touchscreen functionality is implemented in the low level ADC driver.
 * During initialization, this driver creates a kernel thread.  This thread
 * then calls the ADC driver to obtain touchscreen values continously. These
 * values are then passed to the input susbsystem.
 *
 * @ingroup touchscreen
 */

#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/kthread.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/input.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/freezer.h>
#include <linux/regulator/consumer.h>
#include <linux/imx_adc.h>

#define IMX_ADC_TS_NAME	"imx_adc_ts"

struct imx_adc_ts_drvdata {
	struct task_struct *thread;
	struct regulator *regulator;
};

static int imx_adc_ts_thread(void *arg)
{
	struct input_dev *input = arg;
	struct t_touch_screen ts_sample;
	int wait = 0;

	while (!kthread_should_stop()) {
		memset(&ts_sample, 0, sizeof(ts_sample));
		if (0 != imx_adc_get_touch_sample(&ts_sample, !wait))
			continue;

		wait = ts_sample.contact_resistance;
		if (ts_sample.contact_resistance) {
			if (!ts_sample.x_position && !ts_sample.y_position)
				goto ignore_sample;
			input_report_abs(input, ABS_X, ts_sample.x_position);
			input_report_abs(input, ABS_Y, ts_sample.y_position);
			input_report_abs(input, ABS_PRESSURE, 1);
			input_report_key(input, BTN_TOUCH, 1);
		} else {
			input_report_abs(input, ABS_PRESSURE, 0);
			input_report_key(input, BTN_TOUCH, 0);
		}
		input_sync(input);
	ignore_sample:
		msleep(10);
	}

	return 0;
}

static int __devinit imx_adc_ts_probe(struct platform_device *pdev)
{
	struct input_dev *input;
	struct imx_adc_ts_drvdata *drvdata;
	struct imx_adc_ts_platform_data *plat = pdev->dev.platform_data;
	bool should_wakeup = 0;
	int ret;

	if (!is_imx_adc_ready())
		return -ENODEV;

	input = input_allocate_device();
	if (!input) {
		dev_err(&pdev->dev, "not enough memory for input device\n");
		return -ENOMEM;
	}

	drvdata = kzalloc(sizeof(struct imx_adc_ts_drvdata), GFP_KERNEL);
	if (!drvdata) {
		ret = -ENOMEM;
		goto exit_allocate_device;
	}

	input->name = IMX_ADC_TS_NAME;
	input->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
	input->keybit[BIT_WORD(BTN_TOUCH)] |= BIT_MASK(BTN_TOUCH);
	input_set_abs_params(input, ABS_X,
			     100,
			     4000,
			     0, 0);
	input_set_abs_params(input, ABS_Y,
			     100,
			     4000,
			     0, 0);
	input_set_abs_params(input, ABS_PRESSURE,
			     0, 1, 0, 0);

	ret = input_register_device(input);
	if (ret < 0)
		goto exit_allocate_drvdata;

	if (plat) {
		should_wakeup = plat->should_wakeup;

		if (plat->regulator) {
			drvdata->regulator = regulator_get(&pdev->dev,
							   plat->regulator);
			if (IS_ERR(drvdata->regulator)) {
				dev_err(&input->dev,
					"regulator \"%s\" not found\n",
					plat->regulator);
				ret = PTR_ERR(drvdata->regulator);
				goto exit_register_device;
			}
			regulator_enable(drvdata->regulator);
		}
	}

	device_init_wakeup(&pdev->dev, 1);
	device_set_wakeup_enable(&pdev->dev, should_wakeup);
	imx_adc_register_ts(&pdev->dev);

	drvdata->thread = kthread_run(imx_adc_ts_thread, input,
				      IMX_ADC_TS_NAME);
	if (IS_ERR(drvdata->thread)) {
		ret = PTR_ERR(drvdata->thread);
		goto exit_regulator_enable;
	}

	printk(KERN_INFO "i.MX ADC input touchscreen loaded.\n");

	dev_set_drvdata(&input->dev, drvdata);
	platform_set_drvdata(pdev, input);

	return 0;

exit_regulator_enable:
	regulator_disable(drvdata->regulator);
	regulator_put(drvdata->regulator);
	imx_adc_unregister_ts(&pdev->dev);
exit_register_device:
	input_unregister_device(input);
exit_allocate_drvdata:
	kfree(drvdata);
exit_allocate_device:
	input_free_device(input);
	return ret;
}

static int __devexit imx_adc_ts_remove(struct platform_device *pdev)
{
	struct input_dev *input = platform_get_drvdata(pdev);
	struct imx_adc_ts_drvdata *drvdata = dev_get_drvdata(&input->dev);

	kthread_stop(drvdata->thread);
	if (drvdata->regulator) {
		regulator_disable(drvdata->regulator);
		regulator_put(drvdata->regulator);
	}
	imx_adc_unregister_ts(&pdev->dev);
	input_unregister_device(input);
	kfree(drvdata);
	input_free_device(input);
	platform_set_drvdata(pdev, NULL);

	return 0;
}

#ifdef CONFIG_PM
static int imx_adc_ts_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct input_dev *input = platform_get_drvdata(pdev);
	struct imx_adc_ts_drvdata *drvdata = dev_get_drvdata(&input->dev);

	if (drvdata->regulator)
		if (!device_may_wakeup(&pdev->dev))
			regulator_disable(drvdata->regulator);

	return 0;
}

static int imx_adc_ts_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct input_dev *input = platform_get_drvdata(pdev);
	struct imx_adc_ts_drvdata *drvdata = dev_get_drvdata(&input->dev);

	if (drvdata->regulator)
		if (!device_may_wakeup(&pdev->dev))
			regulator_enable(drvdata->regulator);

	return 0;
}

static struct dev_pm_ops imx_adc_ts_ps_ops = {
	.suspend = imx_adc_ts_suspend,
	.resume = imx_adc_ts_resume,
};
#endif

static struct platform_driver imx_adc_ts_driver = {
	.driver = {
		.name = IMX_ADC_TS_NAME,
		.owner = THIS_MODULE,
#ifdef CONFIG_PM
		.pm = &imx_adc_ts_ps_ops,
#endif
	},
	.probe = imx_adc_ts_probe,
	.remove = __devexit_p(imx_adc_ts_remove),
};

static int __init imx_adc_ts_init(void)
{
	return platform_driver_register(&imx_adc_ts_driver);
}

static void __exit imx_adc_ts_exit(void)
{
	platform_driver_unregister(&imx_adc_ts_driver);
}

late_initcall(imx_adc_ts_init);
module_exit(imx_adc_ts_exit);

MODULE_DESCRIPTION("i.MX ADC input touchscreen driver");
MODULE_AUTHOR("Freescale Semiconductor, Inc.");
MODULE_LICENSE("GPL");
